from jacobsjinjatoo import templator as jj2
import os
import yaml
import typer
from typing_extensions import Annotated
from pathlib import Path
from rich import print

from stingeripc import StingerInterface
from typing import Any


def main(
    inname: Annotated[Path, typer.Argument(exists=True, file_okay=True, dir_okay=False, readable=True)],
    outdir: Annotated[Path, typer.Argument(file_okay=False, dir_okay=True, writable=True, readable=True)],
):
    """Generate C++ output for a Stinger interface."""

    with inname.open(mode="r") as f:
        stinger = StingerInterface.from_yaml(f, placeholder="%1%")

    params: dict[str, Any] = {
        "stinger": stinger,
        "client_source_files": [],
        "server_source_files": [],
        "client_header_files": [],
        "server_header_files": [],
    }

    if outdir.is_file():
        raise RuntimeError("Output directory is a file!")

    if not outdir.is_dir():
        outdir.mkdir(parents=True)

    source_code_dir = outdir / "src"
    source_code_dir.mkdir(parents=True, exist_ok=True)

    headers_code_dir = outdir / "include"
    headers_code_dir.mkdir(parents=True, exist_ok=True)

    examples_code_dir = outdir / "examples"
    examples_code_dir.mkdir(parents=True, exist_ok=True)

    cmake_code_dir = outdir / "cmake"
    cmake_code_dir.mkdir(parents=True, exist_ok=True)

    this_file = Path(__file__)
    # The templates live under the package `stingeripc/templates/cpp` next to `tools`.
    # compute that from the tools directory: tools -> parent (stingeripc) -> templates/cpp
    template_dir = (this_file.parent / ".." / "templates" / "cpp").resolve()

    if not template_dir.is_dir():
        raise RuntimeError(f"C++ template directory not found: {template_dir}")

    output_dir = outdir.resolve()
    t = jj2.CodeTemplator(output_dir=output_dir)
    t.add_template_dir(template_dir)

    headers_template_dir = template_dir / "include"
    for fname in os.listdir(headers_template_dir):
        header_code_path = Path("include") / fname
        if fname.endswith(".jinja2"):
            print(f"GENERATING HEADER: {fname}")
            header_dest_path = str(header_code_path)[:-len(".jinja2")]
            if "server" not in header_dest_path:
                params["client_header_files"].append(header_dest_path)
            if "client" not in header_dest_path:
                params["server_header_files"].append(header_dest_path)
            t.render_template(header_code_path, header_dest_path, **params)

    src_template_dir = template_dir / "src"
    for fname in os.listdir(src_template_dir):
        source_code_path = Path("src") / fname
        if fname.endswith(".jinja2"):
            print(f"GENERATING SOURCE: {fname}")
            source_dest_path = str(source_code_path)[:-len(".jinja2")]
            if "server" not in source_dest_path:
                params["client_source_files"].append(source_dest_path)
            if "client" not in source_dest_path:
                params["server_source_files"].append(source_dest_path)
            t.render_template(source_code_path, source_dest_path, **params)

    example_template_dir = template_dir / "examples"
    for fname in os.listdir(example_template_dir):
        example_code_path = Path("examples") / fname
        if fname.endswith(".jinja2"):
            print(f"GENERATING EXAMPLE: {fname}")
            example_dest_path = str(example_code_path)[:-len(".jinja2")]
            t.render_template(example_code_path, example_dest_path, **params)

    print("GENERATING CmakeLists")
    t.render_template("CMakeLists.txt.jinja2", "CMakeLists.txt", **params)

    env = t._get_jinja2_environment()
    cmake_name = env.filters["PascalCase"](stinger.name)
    print(f"Generating CMake configuration with name {cmake_name}")
    t.render_template("cmake/config.cmake.in.jinja2", f"cmake/{cmake_name}-config.cmake.in", **params)


def run():
    typer.run(main)


if __name__ == "__main__":
    run()




from __future__ import annotations
from enum import Enum
import random
from abc import abstractmethod
from typing import Any, Optional, Mapping
from .topic import (
    SignalTopicCreator,
    InterfaceTopicCreator,
    MethodTopicCreator,
    PropertyTopicCreator,
)
from .lang_symb import *
from .args import ArgType, ArgPrimitiveType
from .exceptions import InvalidStingerStructure
from jacobsjinjatoo import stringmanip
from copy import copy
from stevedore import ExtensionManager

YamlArg = Mapping[str, str | bool]
YamlArgList = list[YamlArg]
YamlIfaceEnum = dict[str, str | YamlArgList]
YamlIfaceEnums = dict[str, YamlIfaceEnum]
YamlIfaceProperty = dict[str, str | bool | YamlArgList]

RESTRICTED_NAMES = ["type", "class", "struct", "enum", "list", "map", "set", "optional", "bool", "int", "float", "string", "datetime", "duration", "binary"]


class LanguageSymbolMixin:

    def __init__(self):
        mgr = ExtensionManager(
            namespace="stinger_symbols",
            invoke_on_load=True,
        )
        for ext in mgr:
            domain = ext.name
            symbols = ext.obj.for_model(self.__class__.__name__, self)
            if symbols is not None:
                setattr(self, domain, symbols)

class Arg:

    def __init__(self, name: str, description: Optional[str] = None):
        self._name = name
        self._description = description.strip() if description else None
        self._default_value = None
        self._type: ArgType = ArgType.UNKNOWN
        self._optional: bool = False

    def set_description(self, description: str) -> Arg:
        self._description = description.strip()
        return self

    def try_set_description_from_spec(self, spec: Mapping[str, Any]) -> Arg:
        if "description" in spec and isinstance(spec["description"], str):
            self.set_description(spec["description"])
        return self

    @property
    def name(self) -> str:
        return self._name

    @property
    def arg_type(self) -> ArgType:
        return self._type

    @property
    def description(self) -> str | None:
        return self._description

    @property
    def optional(self) -> bool:
        return self._optional

    @optional.setter
    def optional(self, value: bool):
        self._optional = value

    @property
    def python_type(self) -> str:
        return self.name

    @property
    def python_class(self) -> str:
        return self.python_type

    @property
    def python_local_type(self) -> str:
        return self.python_type.split(".")[-1]

    @property
    def python_annotation(self) -> str:
        return self.python_class

    @property
    def markdown_type(self) -> str:
        """Default markdown representation for an Arg.

        Subclasses may override this to provide richer markdown links.
        """
        return self.python_type

    @property
    def rust_type(self) -> str:
        return self.name

    @property
    def rust_local_type(self) -> str:
        return self.rust_type

    @property
    def cpp_type(self) -> str:
        return stringmanip.upper_camel_case(self.name)
    
    @property
    def cpp_temp_type(self) -> str:
        if self.optional and "optional" not in self.cpp_type:
            return f"std::optional<{self.cpp_type}>"
        return self.cpp_type

    @property
    def cpp_func_param_type(self) -> str:
        if self.optional and "optional" not in self.cpp_type:
            return f"std::optional<{self.cpp_type}>"
        return self.cpp_type

    @classmethod
    def new_arg_from_stinger(
        cls, arg_spec: YamlArg, stinger_spec: Optional[StingerSpec] = None
    ) -> Arg:
        # arg_spec may be an immutable Mapping; copy to mutable dict for validation/mutation
        spec: dict[str, Any]
        if isinstance(arg_spec, dict):
            spec = arg_spec
        else:
            spec = dict(arg_spec)

        if "type" not in spec:
            raise InvalidStingerStructure("No 'type' in arg structure")
        if "name" not in spec:
            raise InvalidStingerStructure("No 'name' in arg structure")
        elif spec["name"] in RESTRICTED_NAMES:
            spec["name"] = f"{spec['name']}_"
        if not isinstance(arg_spec["type"], str):
            raise InvalidStingerStructure("'type' in arg structure must be a string")
        if not isinstance(arg_spec["name"], str):
            raise InvalidStingerStructure("'name' in arg structure must be a string")

        if hasattr(ArgPrimitiveType, spec["type"].upper()):
            arg = ArgPrimitive.new_arg_primitive_from_stinger(spec)
            if opt := spec.get("optional", False):
                assert isinstance(opt, bool), "Optional field must be a boolean"
                arg.optional = opt
            return arg
        else:
            if stinger_spec is None:
                raise RuntimeError(
                    "Need the root StingerSpec when creating an enum or struct Arg"
                )

        if spec["type"] == "enum":
            if "enumName" not in arg_spec:
                raise InvalidStingerStructure(f"Enum args need a 'enumName'")
            if not isinstance(arg_spec["enumName"], str):
                raise InvalidStingerStructure("'enumName' in arg structure must be a string")
            if arg_spec["enumName"] not in stinger_spec.enums:
                raise InvalidStingerStructure(
                    f"Enum arg '{arg_spec['enumName']}' was not found in the list of stinger spec enums"
                )
            enum_arg = ArgEnum(
                spec["name"], stinger_spec.get_interface_enum(spec["enumName"])
            )
            if opt := spec.get("optional", False):
                if not isinstance(opt, bool):
                    raise InvalidStingerStructure("'optional' in arg structure must be a boolean")
                enum_arg.optional = opt
            enum_arg.try_set_description_from_spec(spec)
            return enum_arg

        if spec["type"] == "struct":
            if "structName" not in spec:
                raise InvalidStingerStructure("Struct args need a 'structName'")
            if not isinstance(spec["structName"], str):
                raise InvalidStingerStructure("'structName' in arg structure must be a string")
            if spec["structName"] not in stinger_spec.structs:
                raise InvalidStingerStructure(
                    f"Struct arg '{spec["structName"]}' was not found in the list of stinger spec structs"
                )
            st_arg = ArgStruct(
                spec["name"], stinger_spec.structs[spec["structName"]]
            )
            if opt := spec.get("optional", False):
                if not isinstance(opt, bool):
                    raise InvalidStingerStructure("'optional' in arg structure must be a boolean")
                st_arg.optional = opt
            st_arg.try_set_description_from_spec(spec)
            return st_arg
        
        if spec["type"] == "datetime":
            dt_arg = ArgDateTime(spec["name"])
            if opt := spec.get("optional", False):
                if not isinstance(opt, bool):
                    raise InvalidStingerStructure("'optional' in arg structure must be a boolean")
                dt_arg.optional = opt
            dt_arg.try_set_description_from_spec(spec)
            return dt_arg

        if spec["type"] == "duration":
            dur_arg = ArgDuration(spec["name"])
            if opt := spec.get("optional", False):
                if not isinstance(opt, bool):
                    raise InvalidStingerStructure("'optional' in arg structure must be a boolean")
                dur_arg.optional = opt
            dur_arg.try_set_description_from_spec(spec)
            return dur_arg
        
        if spec["type"] == "binary":
            bin_arg = ArgBinary(spec["name"])
            if opt := spec.get("optional", False):
                if not isinstance(opt, bool):
                    raise InvalidStingerStructure("'optional' in arg structure must be a boolean")
                bin_arg.optional = opt
            bin_arg.try_set_description_from_spec(spec)
            return bin_arg
        
        if spec["type"] == "array":
            if "itemType" not in spec:
                raise InvalidStingerStructure("Array args need an 'itemType'")
            element_arg_spec = copy(spec["itemType"])
            if not isinstance(element_arg_spec, dict):
                raise InvalidStingerStructure("'itemType' in arg structure must be a dict")
            element_arg_spec["name"] = "name_not_used_in_array_element"
            element_arg = Arg.new_arg_from_stinger(element_arg_spec, stinger_spec)
            array_arg = ArgArray(spec["name"], element_arg)
            if opt := spec.get("optional", False):
                if not isinstance(opt, bool):
                    raise InvalidStingerStructure("'optional' in arg structure must be a boolean")
                array_arg.optional = opt
            array_arg.try_set_description_from_spec(spec)
            return array_arg

        raise RuntimeError(f"unknown arg type: {arg_spec['type']}")

    @abstractmethod
    def get_random_example_value(self, lang="python", seed: int = 0): ...


class ArgEnum(Arg, LanguageSymbolMixin):
    
    def __init__(self, name: str, enum: InterfaceEnum, description: Optional[str] = None):
        Arg.__init__(self, name, description)
        LanguageSymbolMixin.__init__(self)
        self._enum = enum
        self._type = ArgType.ENUM

    @property
    def enum(self) -> InterfaceEnum:
        return self._enum

    @property
    def python_type(self) -> str:
        return self._enum.python_type

    @property
    def python_local_type(self) -> str:
        return self._enum.python_local_type

    @property
    def python_class(self) -> str:
        return self._enum.python_type

    @property
    def python_annotation(self) -> str:
        if self.optional:
            return f"Optional[{self._enum.python_type}]"
        return self._enum.python_type

    @property
    def cpp_type(self) -> str:
        if self.optional:
            return f"std::optional<{self._enum.cpp_type}>"
        return self._enum.cpp_type

    @property
    def rust_type(self) -> str:
        if self.optional:
            return f"Option<{self._enum.rust_type}>"
        return self._enum.rust_type

    @property
    def rust_local_type(self) -> str:
        if self.optional:
            return f"Option<{self._enum.rust_local_type}>"
        return self._enum.rust_local_type

    @property
    def cpp_temp_type(self) -> str:
        return self.cpp_type

    @property
    def cpp_data_type(self) -> str:
        return self._enum.cpp_type

    @property
    def cpp_rapidjson_type(self) -> str:
        return ArgPrimitiveType.to_cpp_rapidjson_type_str(ArgPrimitiveType.INTEGER)

    @property
    def markdown_type(self) -> str:
        return f"[Enum {self._enum.class_name}](#enum-{self._enum.class_name})"

    def get_random_example_value(self, lang="python", seed: int = 2) -> str:
        random_state = random.getstate()
        random.seed(seed)
        value = random.choice(self._enum.values)
        if lang == "python":
            retval = f"{self._enum.class_name}.{stringmanip.const_case(value) }"
        elif lang == "c++":
            retval = f"{self._enum.class_name}::{stringmanip.const_case(value)}"
        elif lang == "rust":
            if self.optional:
                retval = f"Some({self._enum.class_name}::{stringmanip.upper_camel_case(value)})"
            else:
                retval = f"{self._enum.class_name}::{stringmanip.upper_camel_case(value)}"
        elif lang == "json":
            retval = str(random.randint(1, len(self._enum.values)))
        random.setstate(random_state)
        return retval

    def __repr__(self) -> str:
        return f"<ArgEnum name={self._name}>"


class ArgPrimitive(Arg, LanguageSymbolMixin):
    
    def __init__(
        self, name: str, arg_type: ArgPrimitiveType, description: Optional[str] = None
    ):
        Arg.__init__(self, name, description)
        LanguageSymbolMixin.__init__(self)
        self._arg_type = arg_type
        self._type = ArgType.PRIMITIVE

    @property
    def type(self) -> ArgPrimitiveType:
        return self._arg_type

    @property
    def python_type(self) -> str:
        return ArgPrimitiveType.to_python_type(self._arg_type)

    @property
    def python_annotation(self) -> str:
        return ArgPrimitiveType.to_python_type(self._arg_type, optional=self._optional)

    @property
    def rust_type(self) -> str:
        return ArgPrimitiveType.to_rust_type(self._arg_type, optional=self._optional)

    @property
    def cpp_type(self) -> str:
        return ArgPrimitiveType.to_cpp_type(self._arg_type, optional=self._optional)

    @property
    def protobuf_type(self) -> str:
        return ArgPrimitiveType.to_protobuf_type(self._arg_type)

    @property
    def cpp_temp_type(self) -> str:
        if self._arg_type == ArgPrimitiveType.STRING:
            if self.optional:
                return "std::optional<std::string>"
            return "std::string"
        else:
            return self.cpp_type

    @property
    def json_type(self) -> str:
        return ArgPrimitiveType.to_json_type(self._arg_type)

    @property
    def cpp_rapidjson_type(self) -> str:
        return ArgPrimitiveType.to_cpp_rapidjson_type_str(self._arg_type)

    def get_random_example_value(
        self, lang="python", seed: int = 2
    ) -> str | float | int | bool | None:
        random_state = random.getstate()
        random.seed(seed)
        retval: str | float | int | bool | None = None
        if self._arg_type == ArgPrimitiveType.BOOLEAN:
            retval = random.choice([True, False])
            if lang != "python":
                retval = str(retval).lower()
        elif self._arg_type == ArgPrimitiveType.FLOAT:
            retval = random.choice([3.14, 1.0, 2.5, 97.9, 1.53])
        elif self._arg_type == ArgPrimitiveType.INTEGER:
            retval = random.choice([42, 1981, 2020, 2022, 1200, 5, 99, 123, 2025, 1955])
        elif self._arg_type == ArgPrimitiveType.STRING:
            retval = random.choice(
                ['"apples"', '"Joe"', '"example"', '"foo"', '"bar"', '"tiger"', '"bear"', '"root beer"']
            )
            if lang == "rust":
                retval = f"{retval}.to_string()"
            if self.optional and lang in ["cpp", "c++"]:
                retval = f'std::make_optional(std::string({retval}))'
        if self.optional and lang == "rust":
            retval = f"Some({retval})"
        random.setstate(random_state)
        return retval

    def __repr__(self) -> str:
        return f"<ArgPrimitive name={self._name} type={ArgPrimitiveType.to_python_type(self.type)}>"

    @classmethod
    def new_arg_primitive_from_stinger(
        cls, arg_spec: Mapping[str, Any]
    ) -> ArgPrimitive:
        if "type" not in arg_spec:
            raise InvalidStingerStructure("No 'type' in arg structure")
        if "name" not in arg_spec:
            raise InvalidStingerStructure("No 'name' in arg structure")
        if not isinstance(arg_spec["type"], str):
            raise InvalidStingerStructure("'type' in arg structure must be a string")
        if not isinstance(arg_spec["name"], str):
            raise InvalidStingerStructure("'name' in arg structure must be a string")

        arg_primitive_type = ArgPrimitiveType.from_string(arg_spec["type"])
        arg: ArgPrimitive = cls(name=arg_spec["name"], arg_type=arg_primitive_type)

        arg.try_set_description_from_spec(arg_spec)
        return arg


class ArgStruct(Arg, LanguageSymbolMixin):

    def __init__(self, name: str, iface_struct: InterfaceStruct):
        Arg.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        assert isinstance(
            iface_struct, InterfaceStruct
        ), f"Passed {iface_struct=} is type {type(iface_struct)} which is not InterfaceStruct"
        self._interface_struct: InterfaceStruct = iface_struct
        self._type = ArgType.STRUCT

    @property
    def struct(self) -> InterfaceStruct:
        return self._interface_struct

    @property
    def members(self) -> list[Arg]:
        return self._interface_struct.members

    @property
    def cpp_type(self) -> str:
        return self._interface_struct.cpp_type

    @property
    def python_type(self) -> str:
        return f"{self._interface_struct.python_local_type}"

    @property
    def python_local_type(self) -> str:
        return self._interface_struct.python_local_type

    @property
    def rust_type(self) -> str:
        if self.optional:
            return f"Option<{self._interface_struct.rust_type}>"
        return self._interface_struct.rust_type

    @property
    def rust_local_type(self) -> str:
        if self.optional:
            return f"Option<{self._interface_struct.rust_local_type}>"
        return self._interface_struct.rust_local_type

    @property
    def rust_temp_type(self) -> str:
        return self._interface_struct.rust_local_type

    @property
    def markdown_type(self) -> str:
        return f"[Struct {self._interface_struct.class_name}](#enum-{self._interface_struct.class_name})"

    @property
    def cpp_rapidjson_type(self) -> str:
        return "Object"

    def get_random_example_value(self, lang="python", seed: int = 2) -> str | None:
        # Build a dict of example values keyed appropriately depending on language.
        example_list: dict[str, str]
        if lang in ["rust", "python"]:
            example_list = {
                stringmanip.snake_case(a.name): str(a.get_random_example_value(lang, seed=seed))
                for a in self.members
            }
        else:
            example_list = {
                a.name: str(a.get_random_example_value(lang, seed=seed))
                for a in self.members
            }
        if lang == "c++":
            return self.cpp_type + "{" + ", ".join(example_list.values()) + "}"
        elif lang == "python":
            init_list = ", ".join([f"{k}={v}" for k, v in example_list.items()])
            return f"{self.python_type}({init_list})"
        elif lang == "rust":
            return "%s%s {%s}%s" % (
                "Some(" if self.optional else "",
                self._interface_struct.rust_type,
                ", ".join([f"{k}: {v}" for k, v in example_list.items()]),
                ")" if self.optional else "",
            )
        elif lang == "json":
            return "{" + ", ".join([f'"{k}": {v}' for k, v in example_list.items()]) + "}"
        return None

    def __str__(self) -> str:
        return f"<ArgStruct name={self.name}>"

    def __repr__(self):
        return f"ArgStruct(name={self.name}, iface_struct={self._interface_struct})"

class ArgDateTime(Arg, LanguageSymbolMixin):
    
    def __init__(self, name: str):
        Arg.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        self._type = ArgType.DATETIME

    @property
    def cpp_type(self) -> str:
        if self.optional:
            return "std::optional<std::chrono::time_point<std::chrono::system_clock>>"
        return "std::chrono::time_point<std::chrono::system_clock>"

    @property
    def cpp_temp_type(self) -> str:
        return self.cpp_type

    @property
    def cpp_rapidjson_type(self) -> str:
        return "String"

    @property
    def python_type(self) -> str:
        return "datetime"

    @property
    def python_local_type(self) -> str:
        return "datetime"
    
    @property
    def python_annotation(self) -> str:
        if self.optional:
            return "Optional[datetime]"
        return "datetime"

    @property
    def rust_type(self) -> str:
        if self.optional:
            return "Option<chrono::DateTime<chrono::Utc>>"
        return "chrono::DateTime<chrono::Utc>"

    @property
    def markdown_type(self) -> str:
        return "[DateTime](#datetime)"

    def get_random_example_value(self, lang="python", seed: int = 2) -> str | None:
        if lang == "python":
            if self.optional and random.choice([True, False, False, False]):
                return "None"
            return f"datetime.now(UTC)"
        elif lang == "rust":
            if self.optional:
                return "Some(chrono::Utc::now())"
            return "chrono::Utc::now()"
        elif lang in ["c++", "cpp"]:
            return "std::chrono::system_clock::now()"
        elif lang == "json":
            return '"1990-07-08T16:20:00Z"'
        return None

    def __str__(self) -> str:
        return f"<ArgDateTime name={self.name}>"

    def __repr__(self):
        return f"ArgDateTime(name={self.name})"

class ArgDuration(Arg, LanguageSymbolMixin):
    
    def __init__(self, name: str):
        Arg.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        self._type = ArgType.DURATION

    @property
    def cpp_type(self) -> str:
        if self.optional:
            return "std::optional<std::chrono::duration<double>>"
        return "std::chrono::duration<double>"

    @property
    def cpp_temp_type(self) -> str:
        return self.cpp_type

    @property
    def python_type(self) -> str:
        return "timedelta"

    @property
    def python_annotation(self) -> str:
        if self.optional:
            return "Optional[timedelta]"
        return "timedelta"

    @property
    def rust_type(self) -> str:
        if self.optional:
            return "Option<chrono::Duration>"
        return "chrono::Duration"

    @property
    def markdown_type(self) -> str:
        return "[Duration](#duration)"

    @property
    def cpp_rapidjson_type(self) -> str:
        return "String"

    def get_random_example_value(self, lang="python", seed: int = 2) -> str | None:
        random_state = random.getstate()
        random.seed(seed)
        retval = None
        if lang == "python":
            if self.optional and random.choice([True, False, False, False]):
                retval = "None"
            else:
                retval = f"timedelta(seconds={random.randint(1, 3600)})"
        elif lang == "rust":
            if self.optional:
                retval = f"Some(chrono::Duration::seconds({random.randint(1, 3600)}))"
            else:
                retval = f"chrono::Duration::seconds({random.randint(1, 3600)})"
        elif lang in ["c++", "cpp"]:
            retval = f"std::chrono::duration<double>({random.randint(1, 3600)})"
        random.setstate(random_state)
        return retval

    def __str__(self) -> str:
        return f"<ArgDuration name={self.name}>"

    def __repr__(self):
        return f"ArgDuration(name={self.name})"
    

class ArgBinary(Arg, LanguageSymbolMixin):
    
    def __init__(self, name: str):
        Arg.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        self._type = ArgType.BINARY

    @property
    def python_type(self) -> str:
        return "bytes"

    @property
    def rust_type(self) -> str:
        if self.optional:
            return "Option<Vec<u8>>"
        return "Vec<u8>"

    @property
    def markdown_type(self) -> str:
        return "[Binary](#binary)"

    @property
    def cpp_rapidjson_type(self) -> str:
        return "String"
    
    @property
    def cpp_type(self) -> str:
        if self.optional:
            return "std::optional<std::vector<uint8_t>>"
        return "std::vector<uint8_t>"
    
    @property
    def cpp_temp_type(self) -> str:
        return self.cpp_type

    def get_random_example_value(self, lang="python", seed: int = 2) -> str | None:
        if lang == "python":
            return f'b"example binary data"'
        elif lang == "rust":
            if self.optional:
                return 'Some(vec![101, 120, 97, 109, 112, 108, 101])'  # "example" in ASCII bytes
            return 'vec![101, 120, 97, 109, 112, 108, 101]'  # "example" in ASCII bytes
        elif lang in ["c++", "cpp"]:
            return 'std::vector<uint8_t>{101, 120, 97, 109, 112, 108, 101}'  # "example" in ASCII bytes
        return None

    def __str__(self) -> str:
        return f"<ArgBinary name={self.name}>"

    def __repr__(self):
        return f"ArgBinary(name={self.name})"
    

class ArgArray(Arg, LanguageSymbolMixin):
    
    def __init__(self, name: str, element_type: Arg):
        Arg.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        self._element = element_type
        self._type = ArgType.ARRAY

    @property
    def element(self) -> Arg:
        return self._element

    @property
    def cpp_type(self) -> str:
        if self.optional:
            return f"std::optional<std::vector<{self.element.cpp_temp_type}>>"
        return f"std::vector<{self.element.cpp_temp_type}>"

    @property
    def python_annotation(self) -> str:
        return f"List[{self.element.python_type}]"

    @property
    def python_type(self) -> str:
        return "list"

    @property
    def rust_type(self) -> str:
        if self.optional:
            return f"Option<Vec<{self.element.rust_type}>>"
        return f"Vec<{self.element.rust_type}>"

    @property
    def markdown_type(self) -> str:
        return f"Array of {self.element.markdown_type}"

    @property
    def cpp_rapidjson_type(self) -> str:
        return "Array"

    def get_random_example_value(self, lang="python", seed: int = 2) -> str | None:
        example_value = self.element.get_random_example_value(lang, seed=seed)
        example_value2 = self.element.get_random_example_value(lang, seed=seed+1)
        example_value3 = self.element.get_random_example_value(lang, seed=seed+2)
        if lang == "python":
            return f"[{example_value}, {example_value2}]"
        elif lang == "rust":
            if self.optional:
                return f"Some(vec![{example_value}, {example_value2}, {example_value3}])"
            return f"vec![{example_value}, {example_value2}]"
        elif lang in ["c++", "cpp"]:
            return f"std::vector<{self.element.cpp_temp_type}>{{{example_value}, {example_value2}}}"
        return None

    def __str__(self) -> str:
        return f"<ArgArray name={self.name} element_type={self.element}>"

    def __repr__(self):
        return f"ArgArray(name={self.name}, element_type={self.element})"


class InterfaceComponent:

    def __init__(self, name: str):
        self._name = name
        self._documentation: Optional[str] = None

    @property
    def name(self) -> str:
        return self._name

    @property
    def documentation(self) -> str | None:
        return self._documentation

    def set_documentation(self, documentation: str) -> InterfaceComponent:
        self._documentation = documentation
        return self

    def try_set_documentation_from_spec(
        self, spec: dict[str, Any]
    ) -> InterfaceComponent:
        if "documentation" in spec and isinstance(spec["documentation"], str):
            self._documentation = spec["documentation"]
        return self


class Signal(InterfaceComponent, LanguageSymbolMixin):
    def __init__(self, topic_creator: SignalTopicCreator, name: str):
        InterfaceComponent.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        self._topic_creator = topic_creator
        self._arg_list: list[Arg] = []

    def add_arg(self, arg: Arg) -> Signal:
        if arg.name in [a.name for a in self._arg_list]:
            raise InvalidStingerStructure(f"An arg named '{arg.name}' has been added.")
        self._arg_list.append(arg)
        return self

    @property
    def arg_list(self) -> list[Arg]:
        return self._arg_list

    @property
    def topic(self) -> str:
        return self._topic_creator.signal_topic(self.name)

    @classmethod
    def new_signal_from_stinger(
        cls,
        topic_creator: SignalTopicCreator,
        name: str,
        signal_spec: dict[str, str],
        stinger_spec: Optional[StingerSpec] = None,
    ) -> "Signal":
        """Alternative constructor from a Stinger signal structure."""
        signal = cls(topic_creator, name)
        if "payload" not in signal_spec:
            raise InvalidStingerStructure("Signal specification must have 'payload'")
        if not isinstance(signal_spec["payload"], list):
            raise InvalidStingerStructure(
                f"Payload must be a list.  It is '{type(signal_spec['payload'])}' "
            )

        for arg_spec in signal_spec["payload"]:
            if "name" not in arg_spec or "type" not in arg_spec:
                raise InvalidStingerStructure("Arg must have name and type.")
            new_arg = Arg.new_arg_from_stinger(arg_spec, stinger_spec)
            signal.add_arg(new_arg)

        signal.try_set_documentation_from_spec(signal_spec)

        return signal


class Method(InterfaceComponent, LanguageSymbolMixin):

    def __init__(self, topic_creator: MethodTopicCreator, name: str):
        InterfaceComponent.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        self._topic_creator = topic_creator
        self._arg_list: list[Arg] = []
        self._return_value: Arg | list[Arg] | None = None
        self._return_arg_list: list[Arg] = []

    def add_arg(self, arg: Arg) -> Method:
        if arg.name in [a.name for a in self._arg_list]:
            raise InvalidStingerStructure(f"An arg named '{arg.name}' has been added.")
        self._arg_list.append(arg)
        return self

    def add_return_value(self, value: Arg) -> Method:
        self._return_arg_list.append(value)
        if self._return_value is None:
            self._return_value = value
        elif isinstance(self._return_value, list):
            if value.name in [a.name for a in self._return_value]:
                raise InvalidStingerStructure(
                    f"A return value named '{value.name}' has been already added."
                )
            self._return_value.append(value)
        elif isinstance(self._return_value, Arg):
            if value.name == self._return_value.name:
                raise InvalidStingerStructure(
                    f"Attempt to add '{value.name}' to return value when it is already been added."
                )
            self._return_value = [self._return_value, value]
        return self

    @property
    def arg_list(self) -> list[Arg]:
        return self._arg_list

    @property
    def return_arg_list(self) -> list[Arg]:
        return self._return_arg_list

    @property
    def return_value(self) -> Arg | list[Arg] | None:
        return self._return_value

    @property
    def return_value_name(self) -> str:
        return f"{self.name} return values"

    @property
    def return_value_cpp_class(self) -> str:
        if self._return_value is None:
            return "void"
        elif isinstance(self._return_value, Arg):
            if isinstance(self._return_value, ArgPrimitive) and self._return_value.type == ArgPrimitiveType.STRING:
                if self._return_value.optional:
                    return "std::optional<std::string>"
                return "std::string"
            elif isinstance(self._return_value, ArgStruct) and self._return_value.optional:
                return f"std::optional<{self._return_value.cpp_type}>"
            return self._return_value.cpp_type
        elif isinstance(self._return_value, list):
            return stringmanip.upper_camel_case(self.return_value_name)

    @property
    def return_value_rust_type(self) -> str:
        if self._return_value is None:
            return "()"
        elif isinstance(self._return_value, Arg):
            return self._return_value.rust_type
        elif isinstance(self._return_value, list):
            return stringmanip.upper_camel_case(self.return_value_name)

    @property
    def return_value_python_type(self):
        if self._return_value is None:
            return "None"
        elif isinstance(self._return_value, Arg):
            return self._return_value.python_type
        elif isinstance(self._return_value, list):
            return (
                f"{stringmanip.upper_camel_case(self.name)}MethodResponse"
            )
        else:
            raise RuntimeError(f"Did not handle return value type for: {self._return_value}")

    @property
    def return_value_python_annotation(self):
        if self._return_value is None:
            return "None"
        elif isinstance(self._return_value, Arg):
            return self._return_value.python_annotation
        elif isinstance(self._return_value, list):
            return (
                f"{stringmanip.upper_camel_case(self.name)}MethodResponse"
            )
        else:
            raise RuntimeError(f"Did not handle return value type for: {self._return_value}")

    @property
    def return_value_python_local_type(self):
        if self._return_value is None:
            return "None"
        elif isinstance(self._return_value, Arg):
            return self._return_value.python_local_type
        elif isinstance(self._return_value, list):
            return stringmanip.upper_camel_case(self.return_value_name)

    @property
    def return_value_property_name(self) -> str:
        if isinstance(self._return_value, Arg):
            return self._return_value.name
        else:
            return self.name

    @property
    def return_value_type(self) -> str | bool:
        if self._return_value is None:
            return False
        elif isinstance(self._return_value, Arg):
            return self._return_value.arg_type.name.lower()
        elif isinstance(self._return_value, list):
            return "multiple"
        raise RuntimeError("Method return value type was not recognized")

    def get_return_value_random_example_value(
        self, lang: str = "python", seed: int = 2
    ):
        if lang == "python":
            if self._return_value is None:
                return "None"
            elif isinstance(self._return_value, Arg):
                return self._return_value.get_random_example_value(lang, seed)
            elif isinstance(self._return_value, list):
                s = ", ".join([f"{a.name}={a.get_random_example_value(lang,seed)}" for a in self._return_value])
                return f"{self.return_value_python_type}({s})"
            else:
                raise RuntimeError(f"Did not handle return value type for: {self._return_value}")
        if lang == "c++" or lang == "cpp":
            if self._return_value is None:
                return "null"
            elif isinstance(self._return_value, Arg):
                return self._return_value.get_random_example_value(lang, seed)
            elif isinstance(self._return_value, list):
                return ", ".join(
                    [
                        str(a.get_random_example_value(lang, seed))
                        for a in self._return_value
                    ]
                )
        raise RuntimeError(f"No random example for return value for {lang}")

    @property
    def topic(self) -> str:
        return self._topic_creator.method_topic(self.name)

    def response_topic(self, iface_name, client_id) -> str:
        return self._topic_creator.method_response_topic(self.name, iface_name, client_id)

    @classmethod
    def new_method_from_stinger(
        cls,
        topic_creator: MethodTopicCreator,
        name: str,
        method_spec: dict[str, str],
        stinger_spec: Optional[StingerSpec] = None,
    ) -> "Method":
        """Alternative constructor from a Stinger method structure."""
        method = cls(topic_creator, name)
        if "arguments" not in method_spec:
            raise InvalidStingerStructure(
                f"Method '{name}' specification must have 'arguments'"
            )
        if not isinstance(method_spec["arguments"], list):
            raise InvalidStingerStructure(
                f"Arguments for '{name}' method must be a list.  It is '{type(method_spec['arguments'])}' "
            )

        for arg_spec in method_spec["arguments"]:
            if "name" not in arg_spec or "type" not in arg_spec:
                raise InvalidStingerStructure("Arg must have name and type.")
            new_arg = Arg.new_arg_from_stinger(arg_spec, stinger_spec)
            method.add_arg(new_arg)

        if "returnValues" in method_spec:
            if not isinstance(method_spec["returnValues"], list):
                raise InvalidStingerStructure(f"ReturnValues must be a list.")

            for arg_spec in method_spec["returnValues"]:
                if "name" not in arg_spec or "type" not in arg_spec:
                    raise InvalidStingerStructure(
                        "Return value must have name and type."
                    )
                new_arg = Arg.new_arg_from_stinger(arg_spec, stinger_spec)
                method.add_return_value(new_arg)

        method.try_set_documentation_from_spec(method_spec)

        return method

class Property(InterfaceComponent, LanguageSymbolMixin):

    def __init__(self, topic_creator: PropertyTopicCreator, name: str):
        InterfaceComponent.__init__(self, name)
        LanguageSymbolMixin.__init__(self)
        self._topic_creator = topic_creator
        self._arg_list: list[Arg] = []
        self._read_only = False

    def add_arg(self, arg: Arg) -> Property:
        if arg.name in [a.name for a in self._arg_list]:
            raise InvalidStingerStructure(f"An arg named '{arg.name}' has been added.")
        self._arg_list.append(arg)
        return self

    @property
    def python_local_type(self) -> str:
        return self.python_class.split(".")[-1]

    @property
    def python_class(self) -> str:
        if len(self._arg_list) == 1:
            return self._arg_list[0].python_class
        else:
            return f"{stringmanip.upper_camel_case(self.name)}Property"

    @property
    def python_annotation(self) -> str:
        if len(self._arg_list) == 1:
            return self._arg_list[0].python_annotation
        else:
            return f"{stringmanip.upper_camel_case(self.name)}Property"

    @property
    def rust_local_type(self) -> str:
        if len(self._arg_list) == 1:
            return self._arg_list[0].rust_local_type
        else:
            return f"{stringmanip.upper_camel_case(self.name)}Property"

    @property
    def rust_type(self) -> str:
        if len(self._arg_list) == 1:
            return self._arg_list[0].rust_type
        else:
            return f"{self.rust_local_type}"

    @property
    def arg_list(self) -> list[Arg]:
        return self._arg_list

    @property
    def value_topic(self) -> str:
        return self._topic_creator.property_value_topic(self.name)

    @property
    def update_topic(self) -> str:
        return self._topic_creator.property_update_topic(self.name)

    def response_topic(self, iface_name, client_id) -> str:
        return self._topic_creator.property_response_topic(self.name, iface_name, client_id)

    @property
    def read_only(self) -> bool:
        return self._read_only

    @classmethod
    def new_method_from_stinger(
        cls,
        topic_creator: PropertyTopicCreator,
        name: str,
        prop_spec: YamlIfaceProperty,
        stinger_spec: Optional[StingerSpec] = None,
    ) -> "Property":
        """Alternative constructor from a Stinger method structure."""
        prop_obj = cls(topic_creator, name)
        if "values" not in prop_spec:
            raise InvalidStingerStructure("Property specification must have 'values'")
        if not isinstance(prop_spec["values"], list):
            raise InvalidStingerStructure(
                f"Values must be a list.  It is '{type(prop_spec['values'])}' "
            )

        for arg_spec in prop_spec["values"]:
            if "name" not in arg_spec or "type" not in arg_spec:
                raise InvalidStingerStructure("Arg must have name and type.")
            new_arg = Arg.new_arg_from_stinger(arg_spec, stinger_spec)
            prop_obj.add_arg(new_arg)

        if r_o := prop_spec.get("readOnly", False):
            if not isinstance(r_o, bool):
                raise InvalidStingerStructure("'readOnly' in property structure must be a boolean")
            prop_obj._read_only = r_o

        prop_obj.try_set_documentation_from_spec(prop_spec)

        return prop_obj

    def __str__(self) -> str:
        return f"Property<name={self.name} values=[{', '.join([a.name for a in self.arg_list])}]>"


class InterfaceEnum(LanguageSymbolMixin):

    def __init__(self, name: str):
        LanguageSymbolMixin.__init__(self)
        self._name = name
        self._values: list[str] = []
        self._value_descriptions: list[str | None] = []
        self._documentation: Optional[str] = None

    def add_value(self, value: str, description: Optional[str] = None):
        self._values.append(value)
        self._value_descriptions.append(description.strip() if description is not None else None)

    @property
    def name(self):
        return self._name

    @property
    def documentation(self) -> str | None:
        return self._documentation

    @property
    def class_name(self):
        return stringmanip.upper_camel_case(self.name)

    @property
    def python_type(self) -> str:
        return f"{stringmanip.upper_camel_case(self.name)}"

    @property
    def python_local_type(self) -> str:
        return f"{stringmanip.upper_camel_case(self.name)}"

    @property
    def rust_local_type(self) -> str:
        return stringmanip.upper_camel_case(self.name)

    @property
    def rust_type(self) -> str:
        return f"{self.rust_local_type}"

    @property
    def cpp_type(self) -> str:
        return stringmanip.upper_camel_case(self.name)

    @property
    def values(self):
        return self._values
    
    def value_description(self, index: int) -> str | None:
        if index < 0 or index >= len(self._value_descriptions):
            return None
        return self._value_descriptions[index]

    @classmethod
    def new_enum_from_stinger(cls, name, enum_spec: YamlIfaceEnum) -> InterfaceEnum:
        ie = cls(name)
        for enum_obj in enum_spec.get("values", []):
            assert isinstance(enum_obj, dict), f"Enum values must be a dicts."
            if "name" in enum_obj and isinstance(enum_obj["name"], str):
                value_description = enum_obj.get("description", None)
                if value_description is not None and not isinstance(value_description, str):
                    raise InvalidStingerStructure(
                        f"InterfaceEnum '{name}' item descriptions must be strings."
                    )
                ie.add_value(enum_obj["name"], description=value_description)
            else:
                raise InvalidStingerStructure(
                    f"InterfaceEnum '{name}' items must have string names."
                )
        doc = enum_spec.get("documentation", None)
        if doc is not None and isinstance(doc, str):
            ie._documentation = doc
        return ie


class InterfaceStruct(LanguageSymbolMixin):

    def __init__(self, name: str):
        LanguageSymbolMixin.__init__(self)
        self._name = name
        self._members: list[Arg] = []
        self._documentation: Optional[str] = None

    def add_member(self, arg: Arg):
        self._members.append(arg)

    @property
    def name(self):
        return self._name

    @property
    def documentation(self) -> str | None:
        return self._documentation

    @property
    def class_name(self):
        return stringmanip.upper_camel_case(self.name)

    @property
    def python_type(self) -> str:
        return f"{stringmanip.upper_camel_case(self.name)}"

    @property
    def python_local_type(self) -> str:
        return stringmanip.upper_camel_case(self.name)

    @property
    def rust_local_type(self) -> str:
        return stringmanip.upper_camel_case(self.name)

    @property
    def rust_type(self) -> str:
        return f"{self.rust_local_type}"

    @property
    def cpp_type(self) -> str:
        return stringmanip.upper_camel_case(self.name)

    @property
    def values(self) -> list[Arg]:
        return self._members

    @property
    def members(self) -> list[Arg]:
        return self._members

    @classmethod
    def new_struct_from_stinger(
        cls,
        name,
        spec: dict[str, str | list[dict[str, str]]],
        stinger_spec: StingerSpec,
    ) -> InterfaceStruct:
        istruct = cls(name)
        for memb in spec.get("members", []):
            if not isinstance(memb, dict):
                raise InvalidStingerStructure("Struct members must be dicts")
            arg = Arg.new_arg_from_stinger(memb, stinger_spec=stinger_spec)
            istruct.add_member(arg)
        documentation = spec.get("documentation", None)
        if documentation is not None and not isinstance(documentation, str):
            raise InvalidStingerStructure("Struct documentation must be a string")
        istruct._documentation = documentation
        return istruct

    def __str__(self) -> str:
        return f"<InterfaceStruct members={[m.name for m in self.members]}>"

    def __repr__(self):
        return f"InterfaceStruct(name={self.name})"


class MqttTransportProtocol(Enum):
    TCP = 0
    WEBSOCKETS = 1


class StingerSpec(LanguageSymbolMixin):

    def __init__(self, topic_creator: InterfaceTopicCreator, interface: dict[str, Any]):
        LanguageSymbolMixin.__init__(self)
        self._topic_creator = topic_creator
        try:
            self._name: str = interface["name"]
            self._version: str = interface["version"]
        except KeyError as e:
            raise InvalidStingerStructure(
                f"Missing interface property in {interface}: {e}"
            )
        except TypeError:
            raise InvalidStingerStructure(
                f"Interface didn't appear to have a correct type"
            )

        self._summary = interface.get("summary")
        self._title = interface.get("title")
        self._documentation = interface.get("documentation")

        self.signals: dict[str, Signal] = {}
        self.properties: dict[str, Property] = {}
        self.methods: dict[str, Method] = {}
        self.enums: dict[str, InterfaceEnum] = {}
        self.structs: dict[str, InterfaceStruct] = {}

    @property
    def method_return_codes(self) -> dict[int, str]:
        return {
            0: "Success",
            1: "Client Error",
            2: "Server Error",
            3: "Transport Error",
            4: "Payload Error",
            5: "Client Serialization Error",
            6: "Client Deserialization Error",
            7: "Server Serialization Error",
            8: "Server Deserialization Error",
            9: "Method Not Found",
            10: "Unauthorized",
            11: "Timeout",
            12: "OutOfSync",
            13: "Unknown Error",
            14: "Not Implemented",
            15: "Service Unavailable",
        }

    @property
    def interface_info_topic(self) -> str:
        return self._topic_creator.interface_info_topic()

    @property
    def summary(self) -> str:
        return self._summary or ""
    
    @property
    def title(self) -> str:
        return self._title or self._name or ""
    
    @property
    def documentation(self) -> str:
        return self._documentation or ""

    def add_signal(self, signal: Signal):
        assert isinstance(signal, Signal)
        self.signals[signal.name] = signal

    def add_method(self, method: Method):
        assert isinstance(method, Method)
        self.methods[method.name] = method

    def add_property(self, prop: Property):
        assert isinstance(prop, Property)
        self.properties[prop.name] = prop

    @property
    def properties_rw(self) -> dict[str, Property]:
        return {k: v for k, v in self.properties.items() if not v.read_only}

    def add_enum(self, interface_enum: InterfaceEnum):
        assert interface_enum is not None
        self.enums[interface_enum.name] = interface_enum

    def add_struct(self, interface_struct: InterfaceStruct):
        assert interface_struct is not None
        self.structs[interface_struct.name] = interface_struct

    @property
    def interface_info(self) -> tuple[str, dict[str, str]]:
        # Return the topic and the minimal info mapping used by AsyncAPI generation
        topic = self.interface_info_topic
        info = {"version": self.version, "title": self.title}
        return (topic, info)

    def uses_enums(self) -> bool:
        return bool(self.enums)

    def get_interface_enum(self, name: str) -> InterfaceEnum:
        if name in self.enums:
            return self.enums[name]
        raise InvalidStingerStructure(f"Enum '{name}' not found in stinger spec")

    @property
    def name(self):
        return self._name

    @property
    def version(self):
        return self._version

    def all_methods_response_topic(self, client_id: str) -> str:
        for method in self.methods.values():
            return method.response_topic(self._name, client_id)
        raise RuntimeError("No methods defined in stinger spec")

    def all_properties_response_topic(self, client_id: str) -> str:
        for prop in self.properties.values():
            return prop.response_topic(self._name, client_id)
        raise RuntimeError("No properties defined in stinger spec")

    @classmethod
    def new_spec_from_stinger(
        cls, topic_creator, stinger: dict[str, Any]
    ) -> StingerSpec:
        if "stingeripc" not in stinger:
            raise InvalidStingerStructure("Missing 'stingeripc' format version")
        if "version" not in stinger["stingeripc"]:
            raise InvalidStingerStructure("Stinger spec version not present")
        if stinger["stingeripc"]["version"] not in ["0.0.7"]:
            raise InvalidStingerStructure(
                f"Unsupported stinger spec version {stinger['stingeripc']['version']}"
            )

        stinger_spec = StingerSpec(topic_creator, stinger["interface"])

        # Enums must come before other components because other components may use enum values.
        try:
            if "enums" in stinger:
                for enum_name, enum_spec in stinger["enums"].items():
                    ie = InterfaceEnum.new_enum_from_stinger(enum_name, enum_spec)
                    assert (
                        ie is not None
                    ), f"Did not create enum from {enum_name} and {enum_spec}"
                    stinger_spec.add_enum(ie)
        except TypeError as e:
            raise InvalidStingerStructure(
                f"Signal specification appears to be invalid: {e}"
            )

        try:
            if "structures" in stinger:
                for struct_name, struct_spec in stinger["structures"].items():
                    istruct = InterfaceStruct.new_struct_from_stinger(
                        struct_name, struct_spec, stinger_spec
                    )
                    assert (
                        istruct is not None
                    ), f"Did not create struct from {struct_name} and {struct_spec}"
                    stinger_spec.add_struct(istruct)
        except TypeError as e:
            raise InvalidStingerStructure(
                f"Struct specification appears to be invalid: {e}"
            )

        try:
            if "signals" in stinger:
                for signal_name, signal_spec in stinger["signals"].items():
                    signal = Signal.new_signal_from_stinger(
                        topic_creator.signal_topic_creator(),
                        signal_name,
                        signal_spec,
                        stinger_spec,
                    )
                    assert (
                        signal is not None
                    ), f"Did not create signal from {signal_name} and {signal_spec}"
                    stinger_spec.add_signal(signal)
        except TypeError as e:
            raise InvalidStingerStructure(
                f"Signal specification appears to be invalid: {e}"
            )

        try:
            if "methods" in stinger:
                for method_name, method_spec in stinger["methods"].items():
                    method = Method.new_method_from_stinger(
                        topic_creator.method_topic_creator(),
                        method_name,
                        method_spec,
                        stinger_spec,
                    )
                    assert (
                        method is not None
                    ), f"Did not create method from {method_name} and {method_spec}"
                    stinger_spec.add_method(method)
        except TypeError as e:
            raise InvalidStingerStructure(
                f"Method specification appears to be invalid: {e}"
            )

        try:
            if "properties" in stinger:
                for prop_name, prop_spec in stinger["properties"].items():
                    prop = Property.new_method_from_stinger(
                        topic_creator.property_topic_creator(),
                        prop_name,
                        prop_spec,
                        stinger_spec,
                    )
                    assert (
                        prop is not None
                    ), f"Did not create property from {prop_name} and {prop_spec}"
                    stinger_spec.add_property(prop)
        except TypeError as e:
            raise InvalidStingerStructure(
                f"Property specification appears to be invalid: {e}"
            )

        return stinger_spec



from pydantic import BaseModel as PydanticBaseModel, Field
from typing import Optional, List, Dict, Literal
from enum import Enum
import importlib
from . import symbols
class BaseModel(PydanticBaseModel):
    @property
    def rust(self):
        # import symbols.rust module dynamically
        rust_mod = importlib.import_module("stingeripc.symbols.rust")
        symbols_class_name = f"{self.__class__.__name__}Symbols"
        symbols_class = getattr(rust_mod, symbols_class_name, None)
        if symbols_class is None:
            raise AttributeError(f"Symbols class '{symbols_class_name}' not found in symbols.rust")
        return symbols_class(self)

    @property
    def python(self):
        py_mod = importlib.import_module("stingeripc.symbols.python")
        symbols_class_name = f"{self.__class__.__name__}Symbols"
        symbols_class = getattr(py_mod, symbols_class_name, None)
        if symbols_class is None:
            raise AttributeError(f"Symbols class '{symbols_class_name}' not found in symbols.python")
        return symbols_class(self)
    
    @property
    def cpp(self):
        cpp_mod = importlib.import_module("stingeripc.symbols.cpp")
        symbols_class_name = f"{self.__class__.__name__}Symbols"
        symbols_class = getattr(cpp_mod, symbols_class_name, None)
        if symbols_class is None:
            raise AttributeError(f"Symbols class '{symbols_class_name}' not found in symbols.cpp")
        return symbols_class(self)

class SpecMetadata(BaseModel):
    version: str = Field(..., description="The version of the specification")

class InterfaceMetadata(BaseModel):
    name: str = Field(..., description="The name of the interface")
    version: str = Field(..., description="The version of the interface", pattern=r"^\d+\.\d+\.\d+$")
    title: Optional[str] = Field(None, description="A short title for the interface")
    summary: Optional[str] = Field(None, description="A brief summary of the interface")
    documentation: Optional[str] = Field(None, description="A brief description of the interface")

class ProtocolType(str, Enum):
    TCP = "tcp"
    WEBSOCKET = "websocket"

class InterfaceEnumValue(BaseModel):
    name: str = Field(..., description="The name of the enum value")
    description: Optional[str] = Field(None, description="A brief description of the enum value")

class InterfaceEnumSpec(BaseModel):
    values: List[InterfaceEnumValue] = Field(..., description="A list of enum values")
    documentation: Optional[str] = Field(None, description="A brief description of the enum")

class InterfaceEnum(BaseModel):
    name: str = Field(..., description="The name of the enum")
    spec: InterfaceEnumSpec = Field(..., description="The specification of the enum")

class GenericArg(BaseModel):
    name: str = Field(..., description="The name of the argument")
    documentation: Optional[str] = Field(None, description="A brief description of the argument")
    optional: Optional[bool] = Field(False, description="Whether the argument is optional")
    index: Optional[int] = Field(None, description="When exporting to a protocol buffer consumer, the index of each argument is required and must be unique")

class IntegerArg(GenericArg):
    type: Literal["integer"] = Field("integer", description="The type of the argument (integer)")

class NumberArg(GenericArg):
    type: Literal["float"] = Field("float", description="The type of the argument (number)")

class StringArg(GenericArg):
    type: Literal["string"] = Field("string", description="The type of the argument (string)")

class BooleanArg(GenericArg):
    type: Literal["boolean"] = Field("boolean", description="The type of the argument (boolean)")

class DateTimeArg(GenericArg):
    type: Literal["datetime"] = Field("datetime", description="The type of the argument (datetime)")

class DurationArg(GenericArg):
    type: Literal["duration"] = Field("duration", description="The type of the argument (duration)")

class BinaryArg(GenericArg):
    type: Literal["binary"] = Field("binary", description="The type of the argument (binary)")

class EnumArg(GenericArg):
    type: Literal["enum"] = Field("enum", description="The type of the argument (enum)")
    enumName: str = Field(..., description="The name of the enum type")

class StructArg(GenericArg):
    type: Literal["struct"] = Field("struct", description="The type of the argument (struct)")
    structName: str = Field(..., description="The name of the struct type")

class ListArg(GenericArg):
    type: Literal["list"] = Field("list", description="The type of the argument (list)")
    itemType: IntegerArg | NumberArg | StringArg | BooleanArg | DateTimeArg | DurationArg | BinaryArg | EnumArg | StructArg = Field(..., description="The type of items in the list")

AnyArg = IntegerArg | NumberArg | StringArg | BooleanArg | DateTimeArg | DurationArg | BinaryArg | EnumArg | StructArg | ListArg

class InterfaceStruct(BaseModel):
    name: str = Field(..., description="The name of the struct field")
    documentation: Optional[str] = Field(None, description="A brief description of the struct field")
    members: List[AnyArg] = Field(..., description="A list of members in the struct")

class ComponentExport(BaseModel):
    version: str = Field(..., description="The version of the component")
    consumers: List[str] = Field(..., description="A list of consumers for the component")

class InterfaceComponent(BaseModel):
    name: str = Field(..., description="The name of the component")
    documentation: Optional[str] = Field(None, description="A brief description of the component")

class InterfaceSignal(InterfaceComponent):
    payload: List[AnyArg] = Field([], description="A list of arguments for the signal")

class InterfaceMethod(InterfaceComponent):
    arguments: List[AnyArg] = Field([], description="A list of arguments for the method")
    returnValues: List[AnyArg] = Field([], description="A list of return values for the method")

class InterfaceProperty(InterfaceComponent):
    values: List[AnyArg] = Field(..., description="A list of values for the property")
    readOnly: Optional[bool] = Field(False, description="Whether the property is read-only")

class StingerSpec(BaseModel):
    stingeripc: SpecMetadata = Field(..., description="Metadata about the Stinger IPC specification")
    interface: InterfaceMetadata = Field(..., description="Metadata about the interface")
    enums: Optional[Dict[str, InterfaceEnum]] = Field(None, description="A dictionary of enums used in the interface")
    structs: Optional[Dict[str, InterfaceStruct]] = Field(None, description="A dictionary of structs used in the interface")
    signals: Optional[Dict[str, InterfaceSignal]] = Field(None, description="A dictionary of signals in the interface")
    methods: Optional[Dict[str, InterfaceMethod]] = Field(None, description="A dictionary of methods in the interface")
    properties: Optional[Dict[str, InterfaceProperty]] = Field(None, description="A dictionary of properties in the interface")
    


"""Configuration models for Stinger IPC."""

from pydantic import BaseModel, Field
from pathlib import Path
from typing import Optional
import tomllib


class PropertyConfig(BaseModel):
    """Configuration options for properties."""
    stinger_owned_values: bool = Field(default=True, description="Whether the stinger-server owns property values")

class PythonConfig(BaseModel):
    """Python-specific configuration options."""
    python37: bool = Field(default=False, description="Generate Python 3.7 compatible code")


class StingerConfig(BaseModel):
    """Root configuration model for Stinger IPC code generation."""
    python: PythonConfig = Field(default_factory=PythonConfig, description="Python generation options")
    properties: PropertyConfig = Field(default_factory=PropertyConfig, description="Property generation options")


def load_config(config_path: Path) -> StingerConfig:
    """
    Load and parse a TOML configuration file.
    
    Args:
        config_path: Path to the TOML configuration file.
        
    Returns:
        A StingerConfig instance with the parsed configuration.
        
    Raises:
        FileNotFoundError: If the config file doesn't exist.
        ValueError: If the TOML is invalid or doesn't match the schema.
    """
    if not config_path.exists():
        raise FileNotFoundError(f"Configuration file not found: {config_path}")
    
    with config_path.open("rb") as f:
        config_dict = tomllib.load(f)
    
    return StingerConfig(**config_dict)



from stingeripc.components import StingerSpec, Signal
from stingeripc.topic import InterfaceTopicCreator

import unittest

class TestSpecCreateManually(unittest.TestCase):
    def setUp(self):
        self.interface = {
            "name": "test_interface",
            "version": "1.2.3",
        }
        itc = InterfaceTopicCreator(self.interface['name'])
        self.spec = StingerSpec(itc, self.interface)
        signal = Signal(itc.signal_topic_creator(), "mySignal")
        self.spec.add_signal(signal)

    def test_create_spec(self):
        self.assertEqual(self.spec.name, self.interface['name'])

    def test_has_signals(self):
        self.assertEqual(len(self.spec.signals), 1)
        self.assertIn("mySignal", self.spec.signals)

class TestSpecCreateFromStructure(unittest.TestCase):
    def setUp(self):
        self.stinger = {
            "stingeripc": {
                "version": "0.0.7"
            },
            "interface": {
                "name": "test_interface",
                "version": "0.0.1",
            },
            "signals": {
                "mySignal": {
                    "payload": [
                        {"name": "one", "type": "integer"},
                        {"name": "two", "type": "string"},
                    ]
                }
            },
            "enums": {
                "milk": [
                    {"name": "one percent"},
                    {"name": "two percent"},
                ]
            }
        }
        itc = InterfaceTopicCreator(self.stinger['interface']['name'])
        self.spec = StingerSpec.new_spec_from_stinger(itc, self.stinger)

    def test_create_spec(self):
        self.assertIsNotNone(self.spec)
        self.assertEqual(self.spec.name, self.stinger['interface']['name'])

    def test_has_signals(self):
        self.assertEqual(len(self.spec.signals), 1)
        self.assertIn("mySignal", self.spec.signals)

    def test_has_enums(self):
        self.assertEqual(len(self.spec.enums), 1)
        self.assertIn("milk", self.spec.enums)



from jacobsjinjatoo import stringmanip

from stingeripc.args import ArgPrimitiveType
from stingeripc.exceptions import InvalidStingerStructure

class ISymbolsProvider:

    def for_model(self, model_class_name:str, model) -> object|None:
        return None

class ModelSymbols:
    
    def __init__(self, model):
        self._model = model
class RustSymbolsProvider(ISymbolsProvider):

    def for_model(self, model_class_name:str, model) -> object|None:
        if model_class_name == "StingerSpec":
            return RustInterfaceSymbols(model)
        return None


class PythonSymbolsProvider(ISymbolsProvider):

    def for_model(self, model_class_name:str, model) -> object|None:
        if model_class_name == "StingerSpec":
            return PythonInterfaceSymbols(model)
        elif model_class_name == "InterfaceStruct":
            return PythonStructSymbols(model)
        elif model_class_name == "Method":
            return PythonMethodSymbols(model)
        return None

class CppSymbolsProvider(ISymbolsProvider):

    def for_model(self, model_class_name:str, model) -> object|None:
        if model_class_name == "StingerSpec":
            return CppInterfaceSymbols(model)
        return None

class PythonSymbols:

    def __init__(self):
        ...

    @property
    def type_definition_module(self) -> str:
        return "interface_types"


class PythonInterfaceSymbols(PythonSymbols):

    def __init__(self, interface):
        super().__init__()
        self._iface = interface

    @property
    def package_name(self):
        s = f"{stringmanip.lower_camel_case(self._iface.name).lower()}ipc"
        return s.replace('__', '_')

    @property
    def client_class_name(self) -> str:
        """ Name of the python class for the interface client."""
        return f"{stringmanip.upper_camel_case(self._iface.name)}Client"

    @property
    def server_class_name(self) -> str:
        """ Name of the python class for the interface server."""
        return f"{stringmanip.upper_camel_case(self._iface.name)}Server"


class PythonStructSymbols(PythonSymbols):

    def __init__(self, iface_struct):
        super().__init__()
        self._iface_struct = iface_struct


class PythonMethodSymbols(PythonSymbols):

    def __init__(self, method):
        super().__init__()
        self._method = method
    
    @property
    def return_value_annotation(self) -> str:
        return self._method.return_value_python_type
    
    @property
    def return_value_local_class(self) -> str:
        return f"{stringmanip.upper_camel_case(self._method.name)}ReturnValue"

    @property
    def return_value_class(self):
        return f"{self.type_definition_module}.{self.return_value_local_class}"

class RustSymbols:
    def __init__(self):
        pass

class RustInterfaceSymbols(RustSymbols):

    def __init__(self, interface):
        super().__init__()
        self._iface = interface

    @property
    def package_name(self) -> str:
        """ Name of the rust package for the interface client."""
        s = f"{stringmanip.snake_case(self._iface.name)}_ipc"
        return s.replace('__', '_')

    @property
    def client_struct_name(self) -> str:
        """ Name of the rust struct for the interface client."""
        return f"{stringmanip.upper_camel_case(self._iface.name)}Client"

    @property
    def server_struct_name(self) -> str:
        """ Name of the struct for the interface server."""
        return f"{stringmanip.upper_camel_case(self._iface.name)}Server"

class CppSymbols:
    def __init__(self):
        pass

class CppInterfaceSymbols(CppSymbols):
    
    def __init__(self, interface):
        super().__init__()
        self._iface = interface

    @property
    def client_class_name(self) -> str:
        return f"{stringmanip.upper_camel_case(self._iface.name)}Client"
    
    @property
    def server_class_name(self) -> str:
        return f"{stringmanip.upper_camel_case(self._iface.name)}Server"

    @property
    def enum_header_file(self) -> str:
        return "enums.hpp"
    
    @property
    def property_struct_header_file(self) -> str:
        return "property_structs.hpp"




from stingeripc.components import Signal, InvalidStingerStructure
from stingeripc.topic import SignalTopicCreator

import unittest


class TestSignalCreateManually(unittest.TestCase):
    def setUp(self):
        stc = SignalTopicCreator("test_interface")
        self.sig = Signal(stc, "mySignal")

    def test_create_signal(self):
        self.assertEqual(self.sig.name, "mySignal")

    def test_topic(self):
        self.assertEqual(self.sig.topic, "test_interface/signal/mySignal")


class TestSignalCreateFromStinger(unittest.TestCase):
    def setUp(self):
        stc = SignalTopicCreator("test_interface")
        stinger = {
            "payload": [
                {"name": "one", "type": "integer"},
                {"name": "two", "type": "string", "description": "This is the second arg"},
            ]
        }
        self.sig = Signal.new_signal_from_stinger(stc, "mySignal", stinger)

    def test_create_signal(self):
        self.assertEqual(self.sig.name, "mySignal")

    def test_signal_args(self):
        self.assertEqual(len(self.sig.arg_list), 2)

    def test_topic(self):
        self.assertEqual(self.sig.topic, "test_interface/signal/mySignal")


class TestSignalRejectsArgsWithSameName(unittest.TestCase):

    def setUp(self):
        self.topic_creator = SignalTopicCreator("test_interface")

    def test_two_args_with_same_name(self):
        signal_spec = {
            "payload": [
                {"name": "one", "type": "integer"},
                {"name": "one", "type": "string", "description": "This is the second arg"},
            ]
        }

        with self.assertRaises(InvalidStingerStructure):
            Signal.new_signal_from_stinger(self.topic_creator, "badSignal", signal_spec)


"""
Provides the functionality needed to create an AsyncAPI service specification from a Stinger file.
"""

import sys
from jacobsjinjatoo import templator as jj2
from jacobsjinjatoo import stringmanip
import os.path
from enum import Enum
from typing import Any
from collections import OrderedDict
from .components import StingerSpec, Arg, ArgPrimitive, ArgEnum, ArgStruct
from .args import ArgType, ArgPrimitiveType


class Direction(Enum):
    SERVER_PUBLISHES = 1
    SERVER_SUBSCRIBES = 2


class SpecType(Enum):
    SERVER = 1
    CLIENT = 2
    LIB = 3


class ObjectSchema:

    def __init__(self):
        self._properties: dict[str, Any] = dict()
        self._required = set()
        self._dependent_schemas = {}

    def add_value_property(
        self, name: str, arg_primitive_type: ArgPrimitiveType, required=True
    ):
        schema = {"type": ArgPrimitiveType.to_json_type(arg_primitive_type)}
        self._properties[name] = schema
        if required:
            self._required.add(name)

    def add_value_dependency(self, name: str, required_on_name: str, required_on_value):
        self._dependent_schemas[name] = {
            "properties": {required_on_name: {"const": required_on_value}}
        }

    def add_const_value_property(
        self, name: str, arg_type: ArgPrimitiveType, const_value, required=True
    ):
        self.add_value_property(name, arg_type, required)
        self._properties[name]["const"] = const_value

    def add_enum_value_property(
        self, name: str, arg_type: ArgPrimitiveType, possible_values, required=True
    ):
        self.add_value_property(name, arg_type, required)
        self._properties[name]["enum"] = possible_values

    def add_reference_property(self, name: str, dollar_ref: str, required=True):
        schema = {"$ref": dollar_ref}
        self._properties[name] = schema
        if required:
            self._required.add(name)

    def to_schema(self) -> dict[str, str | dict[str, Any] | list[str]]:
        props: dict[str, Any] = dict()
        schema: dict[str, str | dict[str, Any] | list[str]] = {
            "type": "object",
            "properties": props,
            "required": sorted(list(self._required)),
        }
        for prop_name, prop_schema in self._properties.items():
            props[prop_name] = prop_schema
        return schema


class Message(object):
    """The information needed to create an AsyncAPI Message structure."""

    def __init__(self, message_name: str, schema: str | None = None):
        self.name = message_name
        self.schema = schema or {"type": "null"}
        self._traits: list[dict[str, Any]] = list()
        self._headers: dict[str, tuple[bool, Any]] = dict()

    def set_schema(self, schema: dict[str, Any]):
        self.schema = schema
        return self

    def set_reference(self, reference):
        return self.set_schema({"$ref": reference})

    def add_trait(self, trait):
        self._traits.append(trait)

    def add_header(self, name: str, schema: dict[str, Any], required: bool = False):
        self._headers[name] = (required, schema)

    def get_message(self) -> dict[str, Any]:
        msg: dict[str, Any] = {
            "name": self.name,
            "payload": self.schema,
        }
        if len(self._traits) > 0:
            msg["traits"] = self._traits
        if len(self._headers) > 0:
            msg["headers"] = OrderedDict(
                {
                    "properties": OrderedDict(),
                    "required": list(),
                }
            )
            for header_name, (required, schema) in self._headers.items():
                msg["headers"]["properties"][header_name] = schema
                if required:
                    msg["headers"]["required"].append(header_name)
        return msg


class Channel(object):
    """The data needed to create an AsyncAPI Channel structure."""

    def __init__(
        self,
        topic: str,
        name: str,
        direction: Direction,
        message_name: str | None = None,
    ):
        self.topic = topic
        self.name = name
        self.direction = direction
        self.message_name = message_name or name
        self.mqtt = {"qos": 1, "retain": False}
        self.description: str | None = None
        self.parameters: dict[str, str] = dict()
        self._operation_traits: list[dict[str, Any]] = list()

    def set_mqtt(self, qos: int, retain: bool):
        self.mqtt = {"qos": qos, "retain": retain}
        return self

    def set_description(self, description: str):
        self.description = description
        return self

    def add_topic_parameters(self, name: str, json_schema_type: str):
        self.parameters[name] = json_schema_type
        return self

    def add_operation_trait(self, trait: dict[str, Any]):
        self._operation_traits.append(trait)
        return self

    def get_operation(
        self, client_type: SpecType, use_common=False
    ) -> dict[str, dict[str, Any]]:
        channel_item: dict[str, dict[str, Any]] = dict()
        op_item: OrderedDict[str, Any] = OrderedDict(
            {
                "operationId": self.name,
                "message": {
                    "$ref": f"{use_common or ''}#/components/messages/{self.message_name}"
                },
            }
        )
        if use_common is not False:
            op_item["traits"] = [
                {"$ref": f"{use_common}#/components/operationTraits/{self.name}"}
            ]
        elif len(self._operation_traits) > 0:
            op_item.update(
                OrderedDict(
                    {
                        "traits": self._operation_traits,
                    }
                )
            )
        if (
            client_type == SpecType.SERVER
            and self.direction == Direction.SERVER_PUBLISHES
        ) or (
            client_type == SpecType.CLIENT
            and self.direction == Direction.SERVER_SUBSCRIBES
        ):
            channel_item.update({"publish": op_item})
        else:
            channel_item.update({"subscribe": op_item})
        if len(self.parameters) > 0:
            params_obj = dict()
            for param_name, param_type in self.parameters.items():
                params_obj[param_name] = {"schema": {"type": param_type}}
            channel_item.update({"parameters": params_obj})
        return channel_item


class Server(object):
    def __init__(self, name: str):
        self.name = name
        self._protocol = "mqtt"
        self._host: str | None = None
        self._port: int | None = None
        self._lwt_topic: str | None = None

    def set_host(self, host: str, port: int):
        self._host = host
        self._port = port
        return self

    def set_lwt_topic(self, topic: str):
        self._lwt_topic = topic
        return self

    @property
    def url(self) -> str:
        return "{}:{}".format(self._host or "{hostname}", self._port or "{port}")

    def get_server(self) -> dict[str, Any]:
        spec: dict[str, Any] = {
            "protocol": self._protocol,
            "protocolVersion": "5",
            "url": self.url,
        }
        if self._lwt_topic is not None:
            spec["bindings"] = OrderedDict(
                {
                    "mqtt": OrderedDict(
                        {
                            "lastWill": OrderedDict(
                                {
                                    "retain": False,
                                    "message": None,
                                    "qos": 1,
                                    "topic": self._lwt_topic,
                                }
                            )
                        }
                    )
                }
            )
        if self._host is None or self._port is None:
            spec["variables"] = {}
        if self._host is None:
            spec["variables"]["hostname"] = {
                "description": "The hosthame or IP address of the MQTT broker."
            }
        if self._port is None:
            spec["variables"]["port"] = {"description": "The port for the MQTT server"}
        return spec


class AsyncApiCreator(object):
    """A class to create a AsyncAPI specification from several AsyncAPI structures.

    It also accepts a Stinger spec for creating all the structures.
    """

    def __init__(self):
        self.info = dict()
        self.asyncapi: OrderedDict[str, Any] = OrderedDict(
            {
                "asyncapi": "2.4.0",
                "id": "",
                "info": OrderedDict(),
                "channels": OrderedDict(),
                "components": OrderedDict(
                    {
                        "operationTraits": OrderedDict(
                            {
                                "methodCall": OrderedDict(
                                    {
                                        "bindings": OrderedDict(
                                            {
                                                "mqtt": OrderedDict(
                                                    {
                                                        "bindingVersion": "0.2.0",
                                                        "qos": 2,
                                                        "retain": False,
                                                    }
                                                ),
                                            }
                                        ),
                                    }
                                ),
                                "methodCallback": OrderedDict(
                                    {
                                        "bindings": OrderedDict(
                                            {
                                                "mqtt": OrderedDict(
                                                    {
                                                        "bindingVersion": "0.2.0",
                                                        "qos": 1,
                                                        "retain": False,
                                                    }
                                                ),
                                            }
                                        ),
                                    }
                                ),
                                "signal": OrderedDict(
                                    {
                                        "bindings": OrderedDict(
                                            {
                                                "mqtt": OrderedDict(
                                                    {
                                                        "bindingVersion": "0.2.0",
                                                        "qos": 2,
                                                        "retain": False,
                                                    }
                                                ),
                                            }
                                        ),
                                    }
                                ),
                            }
                        ),
                        "messageTraits": OrderedDict(
                            {
                                "methodJsonArguments": OrderedDict(
                                    {
                                        "bindings": OrderedDict(
                                            {
                                                "mqtt": OrderedDict(
                                                    {
                                                        "bindingVersion": "0.2.0",
                                                        "contentType": "application/json",
                                                        "correlationData": OrderedDict(
                                                            {
                                                                "type": "string",
                                                                "format": "uuid",
                                                            }
                                                        ),
                                                    }
                                                ),
                                            }
                                        ),
                                    }
                                ),
                                "methodJsonResponse": OrderedDict(
                                    {
                                        "bindings": OrderedDict(
                                            {
                                                "mqtt": OrderedDict(
                                                    {
                                                        "bindingVersion": "0.2.0",
                                                        "contentType": "application/json",
                                                        "correlationData": OrderedDict(
                                                            {
                                                                "type": "string",
                                                                "format": "uuid",
                                                            }
                                                        ),
                                                        "responseTopic": {
                                                            "type": "string",
                                                        },
                                                    }
                                                ),
                                            }
                                        ),
                                    }
                                ),
                                "signalJson": OrderedDict(
                                    {
                                        "bindings": OrderedDict(
                                            {
                                                "mqtt": OrderedDict(
                                                    {
                                                        "bindingVersion": "0.2.0",
                                                        "contentType": "application/json",
                                                    }
                                                ),
                                            }
                                        ),
                                    }
                                ),
                            }
                        ),
                        "messages": OrderedDict(),
                        "schemas": OrderedDict(),
                    }
                ),
            }
        )
        self.channels = []
        self.messages = []
        self.servers = []
        self.name = "interface"

    def add_schema(self, schema_name: str, schema_spec: dict[str, Any]):
        schema_dict: dict[str, Any] = self.asyncapi["components"]["schemas"]
        schema_dict[schema_name] = schema_spec

    def add_channel(self, channel: Channel):
        self.channels.append(channel)

    def add_message(self, message: Message):
        self.messages.append(message)

    def add_server(self, server: Server):
        self.servers.append(server)

    def set_interface_name(self, name):
        self.name = name
        self.asyncapi["id"] = f"urn:stingeripc:{name}"

    def add_to_info(self, key, value):
        self.asyncapi["info"][key] = value

    def get_asyncapi(self, client_type: SpecType, use_common=None):
        spec = self.asyncapi.copy()
        if len(self.servers) > 0:
            spec["servers"] = {}
        for svr in self.servers:
            spec["servers"][svr.name] = svr.get_server()
        for ch in self.channels:
            spec["channels"][ch.topic] = ch.get_operation(
                client_type, use_common or False
            )
        if use_common is None:
            for msg in self.messages:
                spec["components"]["messages"][msg.name] = msg.get_message()
        return spec


class StingerToAsyncApi:

    def __init__(self, stinger: StingerSpec):
        self._asyncapi: AsyncApiCreator = AsyncApiCreator()
        self._stinger: StingerSpec = stinger
        self._convert()

    def _convert(self):
        self._asyncapi.set_interface_name(self._stinger.name)
        self._add_interface_info()
        self._add_servers()
        self._add_enums()
        self._add_signals()
        self._add_methods()
        if len(self._stinger.methods) > 0:
            schema_name = f"stinger_method_return_codes"
            description = [
                f"The stinger_method_return_codes enum has the following values:"
            ]
            accepted_values = []
            for i, enum_value in self._stinger.method_return_codes.items():
                description.append(f"{i} - {enum_value}")
                accepted_values.append(i)
            json_schema = {
                "type": "integer",
                "description": "\n ".join(description),
                "enum": accepted_values,
            }
            self._asyncapi.add_schema(schema_name, json_schema)
        return self

    def _add_interface_info(self):
        topic, info = self._stinger.interface_info
        self._asyncapi.add_to_info("version", info["version"])
        self._asyncapi.add_to_info("title", info["title"])
        ch = Channel(topic, "interfaceInfo", Direction.SERVER_PUBLISHES)
        ch.set_mqtt(qos=1, retain=True)
        self._asyncapi.add_channel(ch)
        msg = Message("interfaceInfo")
        schema = ObjectSchema()
        for k, v in info.items():
            schema.add_const_value_property(k, ArgPrimitiveType.STRING, v)
        msg.set_schema(schema.to_schema())
        self._asyncapi.add_message(msg)

    def _add_servers(self):
        info_topic, _ = self._stinger.interface_info
        for broker_name, broker_spec in self._stinger.brokers.items():
            svr = Server(broker_name)
            if broker_spec.hostname is not None and broker_spec.port is not None:
                svr.set_host(broker_spec.hostname, broker_spec.port)
            svr.set_lwt_topic(info_topic)
            self._asyncapi.add_server(svr)

    def _add_enums(self):
        for enum_name, enum_spec in self._stinger.enums.items():
            schema_name = f"enum_{enum_name}"
            description = [f"The {enum_name} enum has the following values:"]
            accepted_values = []
            for i, enum_value in enumerate(enum_spec.values):
                description.append(f"{i} - {enum_value}")
                accepted_values.append(i)
            json_schema = {
                "type": "integer",
                "description": "\n ".join(description),
                "enum": accepted_values,
            }
            self._asyncapi.add_schema(schema_name, json_schema)

    def _add_signals(self):
        for sig_name, sig_spec in self._stinger.signals.items():
            ch = Channel(sig_spec.topic, sig_name, Direction.SERVER_PUBLISHES)
            ch.add_operation_trait({"$ref": "#/components/operationTraits/signal"})
            self._asyncapi.add_channel(ch)
            msg = Message(sig_name)
            msg.add_trait({"$ref": "#/components/messageTraits/signalJson"})
            schema = ObjectSchema()
            for arg_spec in sig_spec.arg_list:
                if isinstance(arg_spec, ArgPrimitive):
                    schema.add_value_property(arg_spec.name, arg_spec.type)
                elif isinstance(arg_spec, ArgEnum):
                    schema.add_reference_property(
                        arg_spec.name, f"#/components/schemas/enum_{arg_spec.enum.name}"
                    )
            msg.set_schema(schema.to_schema())
            self._asyncapi.add_message(msg)

    def _add_methods(self):
        for method_name, method_spec in self._stinger.methods.items():
            call_ch = Channel(
                method_spec.topic, method_name, Direction.SERVER_SUBSCRIBES
            )
            call_ch.add_operation_trait(
                {"$ref": "#/components/operationTraits/methodCall"}
            )
            self._asyncapi.add_channel(call_ch)
            call_msg = Message(method_name)
            call_msg.add_trait(
                {"$ref": "#/components/messageTraits/methodJsonArguments"}
            )
            call_msg_schema = ObjectSchema()
            for arg_spec in method_spec.arg_list:
                if isinstance(arg_spec, ArgPrimitive):
                    call_msg_schema.add_value_property(arg_spec.name, arg_spec.type)
                elif isinstance(arg_spec, ArgEnum):
                    call_msg_schema.add_reference_property(
                        arg_spec.name, f"#/components/schemas/enum_{arg_spec.name}"
                    )
            call_msg.set_schema(call_msg_schema.to_schema())
            self._asyncapi.add_message(call_msg)

            resp_ch = Channel(
                method_spec.response_topic("{client_id}"),
                f"{method_name}Response",
                Direction.SERVER_PUBLISHES,
            )
            resp_ch.add_operation_trait(
                {"$ref": "#/components/operationTraits/methodCall"}
            )
            self._asyncapi.add_channel(resp_ch)
            resp_msg = Message(f"{method_name}Response")
            resp_msg.add_trait(
                {"$ref": "#/components/messageTraits/methodJsonArguments"}
            )
            resp_msg.add_header(
                "result",
                {"$ref": "#/components/schemas/stinger_method_return_codes"},
                required=True,
            )
            resp_msg.add_header("debug", {"type": "string"}, required=False)
            resp_msg_schema = ObjectSchema()

            def add_arg(arg: Arg):
                if isinstance(arg, ArgPrimitive):
                    resp_msg_schema.add_value_property(
                        arg.name, arg.type, required=True
                    )
                elif isinstance(arg, ArgEnum):
                    resp_msg_schema.add_reference_property(
                        arg.name,
                        f"#/components/schemas/enum_{arg.enum.name}",
                        required=True,
                    )

            if isinstance(method_spec.return_value, ArgStruct):
                for arg_spec in method_spec.return_value.members:
                    add_arg(arg_spec)
                    resp_msg_schema.add_value_dependency(arg_spec.name, "result", 0)
            elif method_spec.return_value is not None:
                assert isinstance(
                    method_spec.return_value, Arg
                ), "Method return value must be a primitive or enum"
                add_arg(method_spec.return_value)
                resp_msg_schema.add_value_dependency(
                    method_spec.return_value_name, "result", 0
                )

            resp_msg.set_schema(resp_msg_schema.to_schema())
            self._asyncapi.add_message(resp_msg)

    def get_asyncapi(self):
        return self._asyncapi.get_asyncapi(SpecType.CLIENT)



from jacobsjinjatoo import templator as jj2
import os
import shutil
import typer
from typing_extensions import Annotated
from typing import Optional, Any
from pathlib import Path
from rich import print
import importlib.resources

import yaml
from stevedore import ExtensionManager
from stingeripc import StingerInterface
from stingeripc.filtering import filter_by_consumer
from stingeripc.config import load_config, StingerConfig
import logging
import yaml
import yamlloader

#logging.basicConfig(level=logging.INFO)

def main(
    inname: Annotated[Path, typer.Argument(exists=True, file_okay=True, dir_okay=False, readable=True)],
    outdir: Annotated[Path, typer.Argument(file_okay=False, dir_okay=True, writable=True, readable=True)],
    language: Annotated[Optional[str], typer.Argument(help="Shortcut for internally provided templates")] = None,
    template_pkg: Annotated[Optional[list[str]], typer.Option(help="Python package(s) containing templates")] = None,
    template_path: Annotated[Optional[list[Path]], typer.Option(help="Filesystem path(s) to template directories")] = None,
    consumer: Annotated[Optional[str], typer.Option("--consumer", help="Consumer name/identifier")] = None,
    config: Annotated[Optional[Path], typer.Option("--config", help="TOML configuration file", exists=True, file_okay=True, dir_okay=False, readable=True)] = None,
):
    """Generate output for a Stinger interface.
    
    At least one of --language, --template-pkg, or --template-path must be provided.
    """
    
    # Validate that at least one template source is provided
    if not language and not template_pkg and not template_path:
        raise typer.BadParameter(
            "At least one of: --language, --template-pkg, or --template-path must be provided"
        )

    config_obj = None
    if config:
        config_obj = load_config(config)
    else:
        config_obj = StingerConfig()

    print(f"🟢   [bold cyan]LOAD:[/bold cyan] {inname}")
    if consumer:
        print(f"💠 CONSUMER {consumer}")
        with inname.open(mode="r") as f:
            yaml_obj = yaml.load(f, Loader=yamlloader.ordereddict.Loader)
            stinger_yaml = filter_by_consumer(yaml_obj, consumer)
            stinger = StingerInterface.from_dict(stinger_yaml)
    else:
        with inname.open(mode="r") as f:
            stinger = StingerInterface.from_yaml(f)

    params: dict[str, Any] = {
        "stinger": stinger, 
        "config": config_obj,
        "consumer": consumer,
    }

    if outdir.is_file():
        raise RuntimeError("Output directory is a file!")
    
    print(f"📁 [bold cyan]OUTPUT:[/bold cyan] {outdir}")
    
    mgr = ExtensionManager(
        namespace="stinger_symbols",
        invoke_on_load=False,
        on_load_failure_callback=lambda mgr, ext, exc: print(f"❌[bold red]EXTFAIL:[/bold red] {ext.name} : {exc}"),
    )
    for ext in mgr:
        print(f"🔌    EXT: {ext.name:7} : {ext.entry_point_target}")

    if not outdir.is_dir():
        print(f"📁  [green]MKDIR[/green]: {outdir}")
        os.makedirs(outdir)

    # Collect all template directories
    template_dirs = []
    
    code_templator = jj2.CodeTemplator(output_dir=outdir)
    web_templator = jj2.WebTemplator(output_dir=outdir)

    # Add template paths
    if template_path:
        for path in template_path:
            if not path.exists():
                raise RuntimeError(f"Template path does not exist: {path}")
            if not path.is_dir():
                raise RuntimeError(f"Template path is not a directory: {path}")
            real_path = path.resolve()
            print(f"[bold cyan]TEMPLATES:[/bold cyan] {real_path}")
            code_templator.add_template_dir(real_path)
            web_templator.add_template_dir(real_path)
            template_dirs.append(real_path)

    # Add template packages
    if template_pkg:
        for pkg_name in template_pkg:
            try:
                print(f"[bold cyan]TEMPLATES:[/bold cyan] {pkg_name}")
                code_templator.add_template_package(pkg_name)
                web_templator.add_template_package(pkg_name)
                pkg_path = importlib.resources.files(pkg_name)
                template_dirs.append(Path(str(pkg_path)))
            except ModuleNotFoundError as e:
                raise RuntimeError(f"Template package not found: {pkg_name}") from e
    
    # Add language-based template directory if language is specified
    if language:
        this_file = Path(__file__)
        this_dir = this_file.parent
        template_dir = (this_dir / "../templates" / language).resolve()
        print(f"[bold cyan]TEMPLATES:[/bold cyan] {template_dir}")
        if not template_dir.exists():
            raise RuntimeError(f"Template directory does not exist: {template_dir}")
        code_templator.add_template_dir(template_dir)
        web_templator.add_template_dir(template_dir)
        template_dirs.append(template_dir)

    def recursive_find_output_files(src_walker: Path, dest_walker: Path) -> list[str]:
        found_files = []
        for entry in os.listdir(src_walker):
            src_entry = src_walker / entry
            dest_entry = dest_walker / entry
            if ('{{' in entry and '}}' in entry) or ('{%' in entry and '%}' in entry):
                rendered_entry_name = code_templator.render_string(entry, **params)
                if len(rendered_entry_name) > 0:
                    dest_entry = dest_walker / rendered_entry_name
                else:
                    continue
            if entry.endswith(".jinja2"):
                dest_path_str = str(dest_entry)[:-len(".jinja2")]
                dest_path = Path(dest_path_str).relative_to(outdir)
                found_files.append(str(dest_path))
            elif src_entry.is_dir():
                found_files.extend(recursive_find_output_files(src_entry, dest_entry))
            elif src_entry.is_file():
                dest_path = Path(dest_entry).relative_to(outdir)
                found_files.append(str(dest_path))
        return found_files
    
    output_file_list = set()
    for template_dir in template_dirs:
        output_file_list.update(recursive_find_output_files(Path(template_dir), Path(outdir)))
    params["output_files"] = list(output_file_list)

    def recursive_render_templates(template_dir, src_walker: Path, dest_walker: Path):
        print(f"🚶   [green]WALK[/green]: {src_walker}")
        for entry in os.listdir(src_walker):
            src_entry = src_walker / entry
            dest_entry = dest_walker / entry
            print(f"🚶  [white]ENTRY[/white]: {src_entry.relative_to(template_dir)}")
            if ('{{' in entry and '}}' in entry) or ('{%' in entry and '%}' in entry):
                rendered_entry_name = code_templator.render_string(entry, **params)
                if len(rendered_entry_name) == 0:
                    print(f"👓   [grey]NAME[/grey]: {entry} -> [excluded]")
                    continue
                dest_entry = dest_walker / rendered_entry_name
                print(f"👓   [grey]NAME[/grey]: {entry} -> {rendered_entry_name}")
            if str(dest_entry).endswith(".jinja2") and '.jinja2' in str(src_entry):
                dest_path_str = str(dest_entry)[:-len(".jinja2")]
                template = str(src_entry.relative_to(template_dir))
                dest_path = Path(dest_path_str).relative_to(outdir)
                print(f"✨  [green]GENER[/green]: {dest_path}")
                if dest_path_str.endswith(".html") or dest_path_str.endswith(".htm"):
                    web_templator.render_template(template, dest_path, **params)
                else:
                    code_templator.render_template(template, dest_path, **params)
            elif src_entry.is_dir():
                print(f"📁  [green]MKDIR[/green]: {dest_entry.resolve()}")
                if not dest_entry.exists():
                    dest_entry.mkdir(parents=True)
                recursive_render_templates(template_dir, src_entry, dest_entry)
            elif src_entry.is_file():
                shutil.copyfile(src_entry, dest_entry)
                print(f"📄   [green]COPY[/green]: {src_entry}")
            else:
                print(f"⚠️    [red]SKIP[/red]: {src_entry} (unknown type)")

    # Process templates from all template directories
    for template_dir in template_dirs:
        src = Path(template_dir)
        recursive_render_templates(template_dir, src, Path(outdir))

def run():
    typer.run(main)


if __name__ == "__main__":
    run()


from stingeripc.exceptions import InvalidStingerStructure
from stingeripc.lang_symb import ISymbolsProvider, ModelSymbols
from stingeripc.args import ArgPrimitiveType
from jacobsjinjatoo import stringmanip

class ProtocolBufferSymbolsProvider(ISymbolsProvider):

    def for_model(self, model_class_name:str, model) -> object|None:
        if model_class_name == "ArgArray":
            return ProtocolBufferArgArraySymbols(model)
        elif model_class_name == "ArgPrimitive":
            return ProtocolBufferArgPrimitiveSymbols(model)
        elif model_class_name == "ArgEnum":
            return ProtocolBufferArgEnumSymbols(model)
        elif model_class_name == "ArgStruct":
            return ProtocolBufferArgStructSymbols(model)
        elif model_class_name == "InterfaceEnum":
            return ProtocolBufferEnumSymbols(model)
        elif model_class_name == "InterfaceStruct":
            return ProtocolBufferStructSymbols(model)
        return None

class ProtocolBufferEnumSymbols(ModelSymbols):

    @property
    def message_name(self) -> str:
        return stringmanip.upper_camel_case(self._model.name)

class ProtocolBufferStructSymbols(ModelSymbols):

    @property
    def message_name(self) -> str:
        return stringmanip.upper_camel_case(self._model.name)

class ProtocolBufferArgArraySymbols(ModelSymbols):

    @property
    def item_type(self) -> str:
        return self._model.element.pb.data_type
    
class ProtocolBufferArgPrimitiveSymbols(ModelSymbols):

    @property
    def data_type(self) -> str:
        if self._model.type == ArgPrimitiveType.BOOLEAN:
            return "bool"
        elif self._model.type == ArgPrimitiveType.INTEGER:
            return "int32"
        elif self._model.type == ArgPrimitiveType.FLOAT:
            return "float"
        elif self._model.type == ArgPrimitiveType.STRING:
            return "string"
        raise InvalidStingerStructure("Unhandled arg type")

class ProtocolBufferArgEnumSymbols(ModelSymbols):

    @property
    def data_type(self) -> str:
        return self._model.enum.pb.message_name

class ProtocolBufferArgStructSymbols(ModelSymbols):

    @property
    def data_type(self) -> str:
        return self._model.struct.pb.message_name




import os
import sys
from wsgiref.validate import validator
from rich import print
from pathlib import Path
import typer
from typing_extensions import Annotated
from typing import Optional
from jacobsjinjatoo import templator as jj2
import jsonschema_rs
import yaml
import yamlloader
from stingeripc.interface import StingerInterface

from . import cpp_generator
from . import generic_generator

app = typer.Typer(help="stinger-ipc generator CLI")


@app.command()
def generate(
    input_file: Annotated[Path, typer.Argument(..., exists=True, file_okay=True, dir_okay=False, readable=True)],
    output_dir: Annotated[Path, typer.Argument(..., file_okay=False, dir_okay=True, writable=True, readable=True)],
    language: Annotated[Optional[str], typer.Option("--language", "-l", help="Language to generate: rust, python, markdown, cpp, web, protobuf")] = None,
    template_pkg: Annotated[Optional[list[str]], typer.Option("--template-pkg", help="Python package(s) containing templates")] = None,
    template_path: Annotated[Optional[list[Path]], typer.Option("--template-path", help="Filesystem path(s) to template directories")] = None,
    consumer: Annotated[Optional[str], typer.Option("--consumer", help="Consumer name/identifier")] = None,
    config: Annotated[Optional[Path], typer.Option("--config", help="TOML configuration file", exists=True, file_okay=True, dir_okay=False, readable=True)] = None,
):
    """Generate code for a Stinger interface.

    INPUT_FILE is the .stinger.yaml file
    OUTPUT_DIR is the directory that will receive generated files
    
    At least one of --language, --template-pkg, or --template-path must be provided.
    """

    # Check if at least one template source is provided
    if not language and not template_pkg and not template_path:
        raise typer.BadParameter(
            "At least one of: --language, --template-pkg, or --template-path must be provided"
        )
    
    # If language is provided, validate and use specialized generators
    if language:
        lang = language.lower()
        if lang not in ("rust", "python", "markdown", "cpp", "web", "protobuf"):
            raise typer.BadParameter("language must be one of: rust, python, markdown, cpp, web, protobuf")

        if lang in ["markdown", "rust", "html", "protobuf", "python", "web"]:
            generic_generator.main(input_file, output_dir, lang, template_pkg, template_path, consumer, config)
        elif lang == "cpp":
            cpp_generator.main(input_file, output_dir)
        
        print(f"Generation for '{lang}' completed.")
    
    # Use generic generator if template-pkg or template-path is provided
    if template_pkg or template_path:
        generic_generator.main(input_file, output_dir, language, template_pkg, template_path, consumer, config)
        print(f"Generation from custom templates completed.")

@app.command()
def validate(input_file: Annotated[Path, typer.Argument(..., exists=True, file_okay=True, dir_okay=False, readable=True)]):
    """Validate a Stinger interface YAML file.

    INPUT_FILE is the .stinger.yaml file
    """
    schema_file = Path(__file__).parent.parent / "schema" / "schema.yaml"
    schema_obj = yaml.load(schema_file.open("r"), Loader=yamlloader.ordereddict.Loader)
    validator = jsonschema_rs.validator_for(schema_obj)
    
    input_obj = yaml.load(input_file.open("r"), Loader=yamlloader.ordereddict.Loader)

    for error in validator.iter_errors(input_obj):
        print(f"Error: {error}")
        print(f"Location: {error.instance_path}")
    sys.exit(0)

@app.command()
def hello():
    print("Hello world")

def run():
    app()

if __name__ == "__main__":
    run()

